@@ -1,4 +1,4 @@
-from typing import Any, Literal, ClassVar, Generator, get_origin, get_args
+from typing import cast, Any, Literal, ClassVar, Generator, get_origin, get_args
 
 import logging
 logger = logging.getLogger(__name__)
@@ -6,11 +6,14 @@ logger = logging.getLogger(__name__)
 import io
 import re
 
+from time import sleep
 import pandas as pd
 
 from collections.abc import Mapping
 from dataclasses import dataclass, field
 
+from pathlib import Path
+
 from inspect import getmembers
 
 from joblib import Parallel, delayed
@@ -20,16 +23,13 @@ from turbodbc import Megabytes
 from turbodbc.connection import Connection
 from turbodbc.cursor import Cursor
 from turbodbc import connect, make_options
-from turbodbc.exceptions import InterfaceError
+from turbodbc.exceptions import InterfaceError, DatabaseError
 
 
 
 
-def is_classvar(x):
-    return isinstance(x, ClassVar)
-
 def is_property(x):
-    return isinstance(x, property)
+    return isinstance(x, config_property)
 
 
 class config_property(property):
@@ -53,9 +53,9 @@ class config_property(property):
                 ):
                     __value = classvar_value
                     break
-            # __value = getattr(__instance.__class__, 'default_' + self.fget.__name__)
         return super().__set__(__instance, __value)
 
+
 @dataclass
 class AbsConfig:
 
@@ -66,8 +66,7 @@ class AbsConfig:
             if get_origin(value.type) is ClassVar:
                 mapping[key] = getattr(cls, key)
         return mapping
-
-
+    
     def property_map(self):
         mapping = {}
         for prop, _ in getmembers(self.__class__, predicate=is_property):
@@ -83,8 +82,6 @@ class AbsConfig:
                     if name in classvar_name:
                         value = classvar_value
                         break
-                
-                # default_value = getattr(self.__class__, 'default_'+name)
                 setattr(self, name, value)
 
 @dataclass
@@ -191,31 +188,23 @@ class TurbODBCOptions(AbsConfig):
     @fetch_wchar_as_char.setter
     def fetch_wchar_as_char(self, fetch_wchar_as_char: bool | None):
         self._fetch_wchar_as_char = fetch_wchar_as_char
-    
-    # def __post_init__(self):
-    #     super().__post_init__()
-    #     print('NOW MAKING OPTIONS')
-    #     print({**self.property_map()})
-    #     print(make_options(**self.property_map()))
-    #     return make_options(**self.property_map())
 
 
 @dataclass
 class SQLConfig(AbsConfig):
     """
-    An obscured declaration is intentional here.
+    NOTE: Mind the `driver` field value. It better be (if available) an explicit ODBC driver with specified version, ex. 'ODBC Driver 17 for SQL Server'. Generic values like 'SQL Server' will seriously limit turbodbc capabilities.
 
-    NOTE: Mind the `driver` field value. It is better be (if available) an ODBC driver with specified version, ex. 'ODBC Driver 17 for SQL Server'. Generic values like 'SQL Server' will seriously limit turbodbc capabilities.
-
-    This is "simplest" way I found so far to 'replace' protected field with a dummy during instantiation.  
+    An obscured attribute declaration is intentional here.
+    This is very ugly, but is "simplest" way I found so far to 'replace' protected field with a dummy during instantiation.  
 
     Ex. with `server` attributes:
 
-    server : is a dummy-field that is used on init and gets added to repr. This gets overwriten with property attribute right away. So it is only here to look pretty.
+    server : is a dummy-field that is used on init and gets added to repr. This gets overwriten with property attribute right away. So it is only here to look "pretty".
 
     _server : a field that holds actual value for an instance
 
-    default_server (prefix `deafult_` is important) - is a classwide default. Attributes that weren't initialized (i.e. initialized with `None`) will borrow from this default. Same if those were reset to `None`.
+    default_server - is classwide default. Attributes initialized as (or reset to) `None` will borrow from this default.
     """
     server: str = field(default=None, compare=False)
     _server: str | None = field(init=False, repr=False, default=None)
@@ -258,50 +247,52 @@ class SQLConfig(AbsConfig):
         self._trusted_connection = trusted_connection
 
     turbodbc_options: Options = field(default=None, compare=False)
-    _turbodbc_options: Options | TurbODBCOptions | None = field(init=False, repr=False, default_factory=TurbODBCOptions)
+    _turbodbc_options: Options | TurbODBCOptions | None = field(init=False, repr=False, default=None)
     @config_property
     def turbodbc_options(self) -> Options | None:
         return self._turbodbc_options
     @turbodbc_options.setter
     def turbodbc_options(self, turbodbc_options: Options | TurbODBCOptions):
-        if type(turbodbc_options) == TurbODBCOptions:
+        if type(turbodbc_options) is TurbODBCOptions:
             turbodbc_options = make_options(**self.property_map())
         self._turbodbc_options = turbodbc_options
     
+
     def __post_init__(self):
         super().__post_init__()
-        if type(self._turbodbc_options) == TurbODBCOptions:
-            self.turbodbc_options = make_options(**self._turbodbc_options.property_map())
-
-
-# @dataclass
-# class SQLConfig(Config):
+        if self._turbodbc_options is None:
+            self._turbodbc_options = TurbODBCOptions()
+        self.turbodbc_options = make_options(**self._turbodbc_options.property_map())
+    
 
     def pre_connect(self) -> None:
-        """Actions that precede connection creation."""
+        """
+        Actions that precede connection creation.  
+        Example:  
+        ```python
+        if time_between(start, end):
+            sleep(30 * 60)
+        ```
+        """
+        logger.debug('Running default pre-connection placeholder procedure.')
         pass
 
     def get_connection(self) -> Connection:
-        # self._pre_connect()
 
-        # turbo_options: Options = ...
-
-        # mapping = {}
-        # for attr in fields(self):
-        #     if not attr.name.startswith('_'):
-        #         attr_name = attr.name
-        #         print(attr_name)
-        #         attr_value = getattr(self, attr.name)
-        #         print(attr_value)
-        #         mapping[attr_name] = attr_value
+        self.pre_connect()
 
         mapping = self.property_map()
+
+        logger.debug('Opening connection.')
         return connect(**mapping)
     
     def get_cursor(self) -> Cursor:
+
         connection = self.get_connection()
+
+        logger.debug('Preparing cursor object.')
         return connection.cursor()
-    
+
 
 @dataclass
 class Query:
@@ -310,66 +301,60 @@ class Query:
 
     ---
     Attributes:  
-    `stmt` :  query statement (alternatively can be a file path, see `__post_init__`)  
+    `stmt` :  query statement (alternatively can be a file path, see `__post_init__`)
     `data` :  input/output data (contextual)
     `conf` :  connection configuration
     `complete` : completion flag
+    `nattempts` : number of retry attempts
+    `fsleep` : sleeptime after failed attempt
 
     ---
-    NOTE
+    NOTE(S)
+    dataclass's default_factory acts funny on inheritance.
     
-    Running multi-statement transactions with turbodbc can be a challenging task.  
+    Running multi-statement transactions with turbodbc can be challenging.  
     
     To name a few examples: 
         - MARS Transactions, see `Query().update_stmt()`
-        - 'USE %DB%;' statement is ignored by 'turbodbc'. Caused by GO?
+        - Non T-SQL statements like GO must be avoided
+        - Very much driver dependent. Driver better be specified explicitly.
     """
 
-    stmt: str = '' 
-    data: pd.DataFrame | None = None
-    conf: SQLConfig = field(default_factory=SQLConfig)
+    stmt: str | Path = '' 
+    data: pd.DataFrame = cast(pd.DataFrame, None)
+    # conf: SQLConfig = field(default_factory=SQLConfig)
+    conf: SQLConfig = cast(SQLConfig, None)
     complete: bool = False
+    nattempts: int = 1
+    fsleep: int = 0
+
 
     def __post_init__(self):
-
+        if self.conf is None:
+            self.conf = SQLConfig()
+        
         try:
             with io.open(self.stmt, mode='rt+', encoding='utf-8') as file:
                 self.stmt = file.read()
-            # logger.debug('Statement is a file path')
-        # except FileNotFoundError:
-        #     # logger.exception(f'File not found: {self.stmt}', exc_info=error)
-        #     raise
+            logger.debug('Statement is a file path.')
         except OSError:
-            # logger.debug('Statement is a string')
-            pass
-        # remove inline comments
-        self.stmt = re.sub(r'--.*?\n', '', self.stmt)
-        logger.debug(f"cleaned up statement:\n{self.stmt}")
-
-        # self.placeholders = re.findall(r'\?', self.stmt)
-        # logger.debug(f"placeholders:\n{self.placeholders}")
+            logger.debug('Statement is a string.')
+        
         self.update_stmt()
     
-   
-    def execute(self, cursor: Cursor | None = None) -> pd.DataFrame | None:
-        """
-        Main wrapper to execute `Query` statement.
-        """
+
+    def execute_once(self, cursor: Cursor | None = None) -> pd.DataFrame:
+        """Main wrapper to execute `Query` statement."""
 
         if not cursor:
             cursor = self.conf.get_cursor()
 
         match self.data:
             case pd.DataFrame():
-                
-                # to correctly convert missing values to SQL NULLs
+                # to correctly convert any missing values to SQL NULLs
                 if self.data.isna().any(axis=None):
                     self.data = self.data.astype('object').where(self.data.notna(), None)
-                
-                # dt_cols = self.data.select_dtypes(include='datetime64[ns]').columns
-                # if not dt_cols.empty:
-                #     self.data[dt_cols] = self.data[dt_cols].stack().dt.strftime('%Y-%m-%d %H:%M:%S').unstack()
-                
+
                 cursor.executemany(self.stmt, self.data.itertuples(index=False))
             case None:
                 cursor.execute(self.stmt)
@@ -383,16 +368,44 @@ class Query:
         except InterfaceError:
             # This error only raised after statement was executed
             # Thus assume it only happens when no result is returned
-            return None
+            return cast(pd.DataFrame, None)
         finally:
             self.complete = True
+    
+    def execute(self, cursor: Cursor | None = None) -> pd.DataFrame:
+        errors = []
+        for attempt in range(self.nattempts):
+            try:
+                return self.execute_once(cursor)
+            except DatabaseError as error:
+                if self.nattempts <= 1:
+                    raise error
+                logger.exception(f'Attempt {attempt + 1}/{self.nattempts} failed.', exc_info=True)
+                errors.append(error)
+                if self.fsleep:
+                    shou = self.fsleep // 3600
+                    smin = (self.fsleep % 3600) // 60
+                    ssec = self.fsleep % 60
+                    logger.info(f'Sleeping for {shou:02d}:{smin:02d}:{ssec:02d}')
+                    sleep(self.fsleep)
+        raise ExceptionGroup(f'All execution retries failed ({self.nattempts=}).', errors)
         
     def update_stmt(self):
         """
+        Additional usecase example:  
         Transactions with Multiple Active Result Sets (MARS) 
-        may return nothing without 'SET NOCOUNT ON;' header.
+        may return nothing without `SET NOCOUNT ON` header.  
+        `self.stmt = 'SET NOCOUNT ON;\n' + self.stmt`
         """
-        self.stmt = 'SET NOCOUNT ON;\n' + self.stmt
+        logger.debug('Running default statement pre-processing.')
+        if not isinstance(self.stmt, str):
+            raise TypeError('Statement should be a valid SQL string by now.')
+        
+        # self.stmt = re.sub(r'--.*?\n', '\n', self.stmt + '\n')[:-2]
+        # logger.debug(f"Cleaned up statement:\n{self.stmt}")
+        
+        # self.placeholders = re.findall(r'\?', self.stmt)
+        # logger.debug(f"Placeholders:\n{self.placeholders}")
 
 
 class Mixin():
@@ -426,15 +439,13 @@ class Mixin():
 @dataclass
 class QuerySequence(Mixin, Mapping):
 
-        
     def run_par(self) -> None:
         """
         Moved to joblib implementation over multiprocessing.
-        Cleaner, simpler and most importantly no issues with pickling.
-        Cons are: extra dependency, joblib logging sucks (latter is yet irrelevant here)
+        Cleaner, simpler and no issues with pickling.
+        Joblib cons are: extra dependency, not as "parallel", logging sucks (latter is yet irrelevant here)
         """
         with Parallel(backend='threading', n_jobs=-1) as parallel:
-        # with Parallel(backend='loky', n_jobs=-1, verbose=True) as parallel:
             dfunc = delayed(Query.execute)
             dqueries = (dfunc(query) for query in self.values())
             parallel(dqueries)
@@ -442,7 +453,7 @@ class QuerySequence(Mixin, Mapping):
 
     def all_confs_equal(self) -> bool:
         """
-        Checks if `conf`s are identical across all `Query` objects in `QuerySequence`.
+        Checks if `conf`s are identical instances across all `Query` objects in `QuerySequence`.
         """
         iter_queries = iter(self.values())
         try:
@@ -454,8 +465,8 @@ class QuerySequence(Mixin, Mapping):
 
     def prep_cursor(self) -> Cursor | None:
         """
-        Creates (and keeps it open) a common `cursor()` instance in case all `conf`s are identical. Relevant for sequential query run.
-
+        Only relevant for sequential query run.  
+        Creates (and keeps open) a common `cursor()` instance in case all `conf`s are identical.   
         Otherwise each `cursor` created separately per `Query` execution.
         """
         if self.all_confs_equal():
@@ -473,7 +484,7 @@ class QuerySequence(Mixin, Mapping):
         cursor = self.prep_cursor()
 
         for key, query in self.items():
-            logger.debug(f"Running query: {key}")
+            logger.info(f'Running query: {key}')
             yield query.execute(cursor)
             
         if cursor:
@@ -488,8 +499,8 @@ class QuerySequence(Mixin, Mapping):
             
             match data:
                 case pd.DataFrame():
-                    logger.debug(f"Data shape: {data.shape}")
+                    logger.info(f'Data shape: {data.shape}')
                 case _:
-                    logger.debug(f"No data returned")
+                    logger.info('No data returned')
             
-            logger.debug(f"{num + 1} / {len(self)} queries done")
+            logger.info(f'{num + 1} / {len(self)} queries complete.')
